@@ -29,7 +29,7 @@
 #ifdef RAR_HIGH_DEBUG
 #define rar_dbgmsg printf
 #else
-static void rar_dbgmsg(const char* fmt,...){}
+static void rar_dbgmsg(const char* fmt,...){(void)fmt;}
 #endif
 
 #define VMCF_OP0             0
@@ -42,6 +42,28 @@ static void rar_dbgmsg(const char* fmt,...){}
 #define VMCF_USEFLAGS       32
 #define VMCF_CHFLAGS        64
 
+#define UINT32(x)  (sizeof(uint32_t)==4 ? (uint32_t)(x):((x)&0xffffffff))
+
+#if WORDS_BIGENDIAN == 0
+#define GET_VALUE(byte_mode,addr) ((byte_mode) ? (*(unsigned char *)(addr)) : UINT32((*(unsigned int *)(addr))))
+#else
+#define GET_VALUE(byte_mode,addr) ((byte_mode) ? (*(unsigned char *)(addr)) : (((unsigned char *)addr)[0] | ((unsigned char *)addr)[1]<<8 | ((unsigned char *)addr)[2]<<16 | ((unsigned char *)addr)[3]<<24))
+#endif
+		
+#if WORDS_BIGENDIAN == 0
+#define SET_VALUE(byte_mode,addr,value) (void)(((byte_mode) ? (*(unsigned char *)(addr)=(value)):(*(uint32_t *)(addr)=((uint32_t)(value)))))
+#else
+#define SET_VALUE(byte_mode,addr,value) rarvm_set_value(byte_mode, (unsigned int *)addr, value);
+#endif
+				
+#define SET_IP(IP)                        \
+  if ((IP)>=(unsigned int)code_size)      \
+    return TRUE;                          \
+  if (--max_ops<=0)                       \
+    return FALSE;                         \
+  cmd=prepared_code+(IP);
+
+
 static uint8_t vm_cmdflags[]=
 {
   /* VM_MOV   */ VMCF_OP2 | VMCF_BYTEMODE                                ,
@@ -86,37 +108,6 @@ static uint8_t vm_cmdflags[]=
   /* VM_PRINT */ VMCF_OP0
 };
 
-#define UINT32(x)  (sizeof(uint32_t)==4 ? (uint32_t)(x):((x)&0xffffffff))
-
-#if WORDS_BIGENDIAN == 0
-#define GET_VALUE(byte_mode,addr) ((byte_mode) ? (*(unsigned char *)(addr)) : UINT32((*(unsigned int *)(addr))))
-#else
-#define GET_VALUE(byte_mode,addr) ((byte_mode) ? (*(unsigned char *)(addr)) : (((unsigned char *)addr)[0] | ((unsigned char *)addr)[1]<<8 | ((unsigned char *)addr)[2]<<16 | ((unsigned char *)addr)[3]<<24))
-#endif
-
-void rarvm_set_value(int byte_mode, unsigned int *addr, unsigned int value)
-{
-	if (byte_mode) {
-		*(unsigned char *)addr=value;
-	} else {
-#if WORDS_BIGENDIAN == 0
-		*(uint32_t *)addr = value;
-#else
-		((unsigned char *)addr)[0]=(unsigned char)value;
-		((unsigned char *)addr)[1]=(unsigned char)(value>>8);
-		((unsigned char *)addr)[2]=(unsigned char)(value>>16);
-		((unsigned char *)addr)[3]=(unsigned char)(value>>24);
-#endif
-	}
-}
-
-		
-#if WORDS_BIGENDIAN == 0
-#define SET_VALUE(byte_mode,addr,value) (void)(((byte_mode) ? (*(unsigned char *)(addr)=(value)):(*(uint32_t *)(addr)=((uint32_t)(value)))))
-#else
-#define SET_VALUE(byte_mode,addr,value) rarvm_set_value(byte_mode, (unsigned int *)addr, value);
-#endif
-
 const uint32_t crc_tab[256]={
 	0x0,        0x77073096, 0xee0e612c, 0x990951ba, 0x76dc419,  0x706af48f, 0xe963a535, 0x9e6495a3,
 	0xedb8832,  0x79dcb8a4, 0xe0d5e91e, 0x97d2d988, 0x9b64c2b,  0x7eb17cbd, 0xe7b82d07, 0x90bf1d91,
@@ -152,14 +143,30 @@ const uint32_t crc_tab[256]={
 	0xb3667a2e, 0xc4614ab8, 0x5d681b02, 0x2a6f2b94, 0xb40bbe37, 0xc30c8ea1, 0x5a05df1b, 0x2d02ef8d
 };
 
+void rarvm_set_value(int byte_mode, unsigned int *addr, unsigned int value)
+{
+	if (byte_mode) {
+		*(unsigned char *)addr=value;
+	} else {
+#if WORDS_BIGENDIAN == 0
+		*(uint32_t *)addr = value;
+#else
+		((unsigned char *)addr)[0]=(unsigned char)value;
+		((unsigned char *)addr)[1]=(unsigned char)(value>>8);
+		((unsigned char *)addr)[2]=(unsigned char)(value>>16);
+		((unsigned char *)addr)[3]=(unsigned char)(value>>24);
+#endif
+	}
+}
+
 uint32_t rar_crc(uint32_t start_crc, void *addr, uint32_t size)
 {
 	unsigned char *data;
-	int i;
+	uint32_t i;
 
 	data = addr;
 #if WORDS_BIGENDIAN == 0
-	while (size > 0 && ((long)data & 7))
+	while (size > 0 && ((size_t)data & 7))
 	{
 		start_crc = crc_tab[(unsigned char)(start_crc^data[0])]^(start_crc>>8);
 		size--;
@@ -213,14 +220,20 @@ void rarvm_addbits(rarvm_input_t *rarvm_input, int bits)
 
 unsigned int rarvm_getbits(rarvm_input_t *rarvm_input)
 {
-	unsigned int bit_field;
+        unsigned int bit_field = 0;
 
-	bit_field = (unsigned int) rarvm_input->in_buf[rarvm_input->in_addr] << 16;
-	bit_field |= (unsigned int) rarvm_input->in_buf[rarvm_input->in_addr+1] << 8;
-	bit_field |= (unsigned int) rarvm_input->in_buf[rarvm_input->in_addr+2];
-	bit_field >>= (8-rarvm_input->in_bit);
+        if (rarvm_input->in_addr < rarvm_input->buf_size) {
+            bit_field = (unsigned int) rarvm_input->in_buf[rarvm_input->in_addr] << 16;
+            if (rarvm_input->in_addr+1 < rarvm_input->buf_size) {
+                bit_field |= (unsigned int) rarvm_input->in_buf[rarvm_input->in_addr+1] << 8;
+                if (rarvm_input->in_addr+2 < rarvm_input->buf_size) {
+                    bit_field |= (unsigned int) rarvm_input->in_buf[rarvm_input->in_addr+2];
+                }
+            }
+        }
+        bit_field >>= (8-rarvm_input->in_bit);
 
-	return (bit_field & 0xffff);
+        return (bit_field & 0xffff);
 }
 
 unsigned int rarvm_read_data(rarvm_input_t *rarvm_input)
@@ -264,7 +277,7 @@ unsigned int rarvm_read_data(rarvm_input_t *rarvm_input)
 static rarvm_standard_filters_t is_standard_filter(unsigned char *code, int code_size)
 {
 	uint32_t code_crc;
-	int i;
+	size_t i;
 
 	struct standard_filter_signature
 	{
@@ -308,10 +321,10 @@ static unsigned int *rarvm_get_operand(rarvm_data_t *rarvm_data,
 	}
 }
 
-static unsigned int filter_itanium_getbits(unsigned char *data, int bit_pos, int bit_count)
+static unsigned int filter_itanium_getbits(unsigned char *data, unsigned int bit_pos, unsigned int bit_count)
 {
-	int in_addr=bit_pos/8;
-	int in_bit=bit_pos&7;
+	unsigned int in_addr=bit_pos/8;
+	unsigned int in_bit=bit_pos&7;
 	unsigned int bit_field=(unsigned int)data[in_addr++];
 	bit_field|=(unsigned int)data[in_addr++] << 8;
 	bit_field|=(unsigned int)data[in_addr++] << 16;
@@ -320,10 +333,10 @@ static unsigned int filter_itanium_getbits(unsigned char *data, int bit_pos, int
 	return(bit_field & (0xffffffff>>(32-bit_count)));
 }
 
-static void filter_itanium_setbits(unsigned char *data, unsigned int bit_field, int bit_pos, int bit_count)
+static void filter_itanium_setbits(unsigned char *data, unsigned int bit_field, unsigned int bit_pos, unsigned int bit_count)
 {
-	int i, in_addr=bit_pos/8;
-	int in_bit=bit_pos&7;
+	unsigned int i, in_addr=bit_pos/8;
+	unsigned int in_bit=bit_pos&7;
 	unsigned int and_mask=0xffffffff>>(32-bit_count);
 	and_mask=~(and_mask<<in_bit);
 
@@ -340,11 +353,12 @@ static void filter_itanium_setbits(unsigned char *data, unsigned int bit_field,
 static void execute_standard_filter(rarvm_data_t *rarvm_data, rarvm_standard_filters_t filter_type)
 {
 	unsigned char *data, cmp_byte2, cur_byte, *src_data, *dest_data;
-	int i, j, data_size, channels, src_pos, dest_pos, border, width, PosR;
-	int op_type, cur_channel, byte_count, start_pos, pa, pb, pc;
+	unsigned int i, j, data_size, channels, src_pos, dest_pos, border, width, PosR;
+	unsigned int op_type, cur_channel, byte_count, start_pos;
+	int pa, pb, pc;
 	unsigned int file_offset, cur_pos, predicted;
-	int32_t offset, addr;
-	const int file_size=0x1000000;
+	uint32_t offset, addr;
+	const unsigned int file_size=0x1000000;
 
 	switch(filter_type) {
 	case VMSF_E8:
@@ -353,7 +367,7 @@ static void execute_standard_filter(rarvm_data_t *rarvm_data, rarvm_standard_fil
 		data_size = rarvm_data->R[4];
 		file_offset = rarvm_data->R[6];
 
-		if (((unsigned int)data_size >= VM_GLOBALMEMADDR) || (data_size < 4)) {
+		if ((data_size > VM_GLOBALMEMADDR) || (data_size < 4)) {
 			break;
 		}
 
@@ -364,12 +378,14 @@ static void execute_standard_filter(rarvm_data_t *rarvm_data, rarvm_standard_fil
 			if (cur_byte==0xe8 || cur_byte==cmp_byte2) {
 				offset = cur_pos+file_offset;
 				addr = GET_VALUE(FALSE, data);
-				if (addr < 0) {
-					if (addr+offset >=0 ) {
+				// We check 0x80000000 bit instead of '< 0' comparison
+				// not assuming int32 presence or uint size and endianness.
+				if ((addr & 0x80000000)!=0) {              // addr<0
+					if (((addr+offset) & 0x80000000)==0) {   // addr+offset>=0
 						SET_VALUE(FALSE, data, addr+file_size);
 					}
 				} else {
-					if (addr<file_size) {
+					if (((addr-file_size) & 0x80000000)!=0) { // addr<file_size
 						SET_VALUE(FALSE, data, addr-offset);
 					}
 				}
@@ -383,7 +399,7 @@ static void execute_standard_filter(rarvm_data_t *rarvm_data, rarvm_standard_fil
 		data_size = rarvm_data->R[4];
 		file_offset = rarvm_data->R[6];
 		
-		if (((unsigned int)data_size >= VM_GLOBALMEMADDR) || (data_size < 21)) {
+		if ((data_size > VM_GLOBALMEMADDR) || (data_size < 21)) {
 			break;
 		}
 		
@@ -426,7 +442,7 @@ static void execute_standard_filter(rarvm_data_t *rarvm_data, rarvm_standard_fil
 		border = data_size*2;
 		
 		SET_VALUE(FALSE, &rarvm_data->mem[VM_GLOBALMEMADDR+0x20], data_size);
-		if ((unsigned int)data_size >= VM_GLOBALMEMADDR/2) {
+		if (data_size > VM_GLOBALMEMADDR/2 || channels > 1024 || channels == 0) {
 			break;
 		}
 		for (cur_channel=0 ; cur_channel < channels ; cur_channel++) {
@@ -437,7 +453,7 @@ static void execute_standard_filter(rarvm_data_t *rarvm_data, rarvm_standard_fil
 		}
 		break;
 	case VMSF_RGB: {
-		const int channels=3;
+		const unsigned int channels=3;
 		data_size = rarvm_data->R[4];
 		width = rarvm_data->R[0] - 3;
 		PosR = rarvm_data->R[1];
@@ -445,15 +461,14 @@ static void execute_standard_filter(rarvm_data_t *rarvm_data, rarvm_standard_fil
 		dest_data = src_data + data_size;
 		
 		SET_VALUE(FALSE, &rarvm_data->mem[VM_GLOBALMEMADDR+0x20], data_size);
-		if ((unsigned int)data_size >= VM_GLOBALMEMADDR/2) {
+		if (data_size > VM_GLOBALMEMADDR/2 || data_size < 3 || width > data_size || PosR > 2) {
 			break;
 		}
 		for (cur_channel=0 ; cur_channel < channels; cur_channel++) {
 			unsigned int prev_byte = 0;
 			for (i=cur_channel ; i<data_size ; i+=channels) {
-				int upper_pos=i-width;
-				if (upper_pos >= 3) {
-					unsigned char *upper_data = dest_data+upper_pos;
+				if (i >= width+3) {
+					unsigned char *upper_data = dest_data+i-width;
 					unsigned int upper_byte = *upper_data;
 					unsigned int upper_left_byte = *(upper_data-3);
 					predicted = prev_byte+upper_byte-upper_left_byte;
@@ -483,13 +498,14 @@ static void execute_standard_filter(rarvm_data_t *rarvm_data, rarvm_standard_fil
 		break;
 	}
 	case VMSF_AUDIO: {
-		int channels=rarvm_data->R[0];
+		unsigned int channels=rarvm_data->R[0];
 		data_size = rarvm_data->R[4];
 		src_data = rarvm_data->mem;
 		dest_data = src_data + data_size;
 		
 		SET_VALUE(FALSE, &rarvm_data->mem[VM_GLOBALMEMADDR+0x20], data_size);
-		if ((unsigned int)data_size >= VM_GLOBALMEMADDR/2) {
+		// In fact, audio channels never exceed 4.
+		if (data_size > VM_GLOBALMEMADDR/2 || channels > 128 || channels == 0) {
 			break;
 		}
 		for (cur_channel=0 ; cur_channel < channels ; cur_channel++) {
@@ -550,7 +566,7 @@ static void execute_standard_filter(rarvm_data_t *rarvm_data, rarvm_standard_fil
 		data_size = rarvm_data->R[4];
 		src_pos = 0;
 		dest_pos = data_size;
-		if ((unsigned int)data_size >= VM_GLOBALMEMADDR/2) {
+		if (data_size > VM_GLOBALMEMADDR/2) {
 			break;
 		}
 		while (src_pos < data_size) {
@@ -567,13 +583,6 @@ static void execute_standard_filter(rarvm_data_t *rarvm_data, rarvm_standard_fil
 		break;
 	}
 }
-				
-#define SET_IP(IP)                      \
-  if ((IP)>=code_size)                   \
-    return TRUE;                       \
-  if (--max_ops<=0)                  \
-    return FALSE;                      \
-  cmd=prepared_code+(IP);
 
 static int rarvm_execute_code(rarvm_data_t *rarvm_data,
 		struct rarvm_prepared_command *prepared_code, int code_size)
